"""Demonstration of the cone_search function."""

# Astropy tools
import astropy.units as u

# For plots
import matplotlib.pyplot as plt
from astropy.coordinates import Angle, Latitude, Longitude, SkyCoord

# Moc and HEALPix tools
from cdshealpix import cone_search
from mocpy import MOC, WCS

max_depth = 10

ipix, depth, fully_covered = cone_search(
    lon=Longitude(0, u.deg), lat=Latitude(0, u.deg), radius=10 * u.deg, depth=max_depth
)


moc = MOC.from_healpix_cells(ipix, depth, max_depth)
# Plot the MOC using matpl/Wordotlib


fig = plt.figure(111, figsize=(10, 10))
# Define a astropy WCS from the mocpy.WCS class
with WCS(
    fig,
    fov=30 * u.deg,
    center=SkyCoord(0, 0, unit="deg", frame="icrs"),
    coordsys="icrs",
    rotation=Angle(0, u.degree),
    projection="AIT",
) as wcs:
    ax = fig.add_subplot(1, 1, 1, projection=wcs)
    # Call fill with a matplotlib axe and the `~astropy.wcs.WCS` wcs object.
    moc.fill(ax=ax, wcs=wcs, alpha=0.5, fill=True, color="green")
    # Draw the perimeter of the MOC in black
    moc.border(ax=ax, wcs=wcs, alpha=0.5, color="black")
plt.xlabel("ra")
plt.ylabel("dec")
plt.title("Cone search")
plt.grid(color="black", linestyle="dotted")
plt.show()



# Astropy tools
import astropy.units as u
from astropy.coordinates import Latitude, Longitude, SkyCoord

import numpy as np
import pytest

from ..ring.healpix import (
    healpix_to_lonlat,
    healpix_to_xy,
    lonlat_to_healpix,
    skycoord_to_healpix,
    vertices,
)


@pytest.mark.parametrize("size", [1, 10, 100, 1000, 10000])
def test_lonlat_to_healpix(size):
    depth = np.random.randint(30)
    nside = 1 << depth

    lon = Longitude(np.random.rand(size) * 360, u.deg)
    lat = Latitude(np.random.rand(size) * 180 - 90, u.deg)

    ipixels, dx, dy = lonlat_to_healpix(
        lon=lon, lat=lat, nside=nside, return_offsets=True
    )
    ipixels, dx, dy = skycoord_to_healpix(
        SkyCoord(lon, lat, frame="icrs"),
        nside=nside,
        return_offsets=True,
    )

    npix = 12 * nside**2
    assert ((ipixels >= 0) & (ipixels < npix)).all()
    assert ((dx >= 0) & (dx <= 1)).all()


@pytest.mark.parametrize(
    "lon, lat, expected_ipix",
    [
        (5 * u.deg, 5 * u.deg, 12),
        (180 * u.deg, 5 * u.deg, 16),
        (179 * u.deg, -88 * u.deg, 45),
    ],
)
def test_lonlat_to_healpix_accurate(lon, lat, expected_ipix):
    nside = 2
    ipixels, dx, dy = lonlat_to_healpix(
        lon=Longitude(lon), lat=Latitude(lat), nside=nside, return_offsets=True
    )

    assert ipixels == expected_ipix
    assert ((dx >= 0) & (dx <= 1)).all()
    assert ((dy >= 0) & (dy <= 1)).all()


@pytest.mark.parametrize("size", [1, 10, 100, 1000, 10000])
def test_healpix_to_lonlat(size):
    depth = np.random.randint(30, dtype=np.uint32)
    nside = 1 << depth

    ipixels = np.random.randint(12 * nside * nside, size=size, dtype=np.uint64)
    lon, lat = healpix_to_lonlat(ipix=ipixels, nside=nside)
    assert lon.shape == lat.shape


@pytest.mark.parametrize("nside", np.arange(start=1, stop=11))
def test_healpix_vs_lonlat(nside):
    size = 1000
    ipixels = np.random.randint(12 * nside * nside, size=size, dtype=np.uint64)
    lon, lat = healpix_to_lonlat(ipix=ipixels, nside=nside, dx=0.5, dy=0.5)

    ipixels_result, dx, dy = lonlat_to_healpix(
        lon=lon, lat=lat, nside=nside, return_offsets=True
    )
    assert (ipixels == ipixels_result).all()


@pytest.mark.parametrize("size", [1, 10, 100, 1000, 10000, 100000])
def test_healpix_to_xy_robust(size):
    depth = np.random.randint(30, dtype=np.uint32)
    nside = 1 << depth

    ipixels = np.random.randint(
        low=0, high=12 * nside * nside, size=size, dtype=np.uint64
    )
    x, y = healpix_to_xy(ipix=ipixels, nside=nside)
    assert x.shape == y.shape


@pytest.mark.parametrize(
    "ipix, depth, expected_x, expected_y",
    [
        (
            np.arange(12),
            0,
            np.array(
                [1.0, 3.0, 5.0, 7.0, 0.0, 2.0, 4.0, 6.0, 1.0, 3.0, 5.0, 7.0],
                dtype=np.float64,
            ),
            np.array(
                [1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, -1, -1, -1, -1],
                dtype=np.float64,
            ),
        ),
        ([], 0, np.array([], dtype=np.float64), np.array([], dtype=np.float64)),
        (0, 0, np.array([1.0], dtype=np.float64), np.array([1.0], dtype=np.float64)),
    ],
)
def test_healpix_to_xy(ipix, depth, expected_x, expected_y):
    x, y = healpix_to_xy(ipix, 1 << depth)

    assert (x == expected_x).all()
    assert (y == expected_y).all()


@pytest.mark.parametrize("size", [1, 10, 100, 1000, 10000, 100000])
def test_vertices_lonlat(size):
    depth = np.random.randint(30)
    nside = 1 << depth

    ipixels = np.random.randint(12 * (nside**2), size=size, dtype=np.uint64)

    lon, lat = vertices(ipix=ipixels, nside=nside)
    assert lon.shape == lat.shape
    assert lon.shape == (size, 4)



# Astropy tools
import astropy.units as u
import astropy_healpix
from astropy.coordinates import Latitude, Longitude

import numpy as np
import pytest

from ..nested.healpix import (
    cone_search,
    healpix_to_lonlat,
    healpix_to_skycoord,
    lonlat_to_healpix,
    neighbours,
    vertices,
)


@pytest.mark.benchmark(group="lonlat_to_healpix_broadcast")
def test_lonlat_to_healpix_broadcast(benchmark):
    size = 10000
    depth = np.random.randint(low=0, high=29, size=10)

    lon = Longitude(np.random.rand(size) * 360, u.deg)
    lat = Latitude(np.random.rand(size) * 180 - 90, u.deg)

    benchmark(
        lonlat_to_healpix,
        lon=lon[:, np.newaxis],
        lat=lat[:, np.newaxis],
        depth=depth[np.newaxis, :],
        num_threads=1,
    )


@pytest.mark.benchmark(group="lonlat_to_healpix_broadcast")
def test_lonlat_to_healpix_broadcast_multithread(benchmark):
    size = 10000
    depth = np.random.randint(low=0, high=29, size=10)

    lon = Longitude(np.random.rand(size) * 360, u.deg)
    lat = Latitude(np.random.rand(size) * 180 - 90, u.deg)

    benchmark(
        lonlat_to_healpix,
        lon=lon[:, np.newaxis],
        lat=lat[:, np.newaxis],
        depth=depth[np.newaxis, :],
    )


@pytest.mark.benchmark(group="lonlat_to_healpix_broadcast")
def test_lonlat_to_healpix_astropy_broadcast(benchmark):
    depth = np.random.randint(low=0, high=29, size=10)
    nside = 2**depth

    size = 10000
    lon = np.random.rand(size) * 360 * u.deg
    lat = (np.random.rand(size) * 180 - 90) * u.deg

    benchmark(
        astropy_healpix.core.lonlat_to_healpix,
        lon=lon[:, np.newaxis],
        lat=lat[:, np.newaxis],
        nside=nside[np.newaxis, :],
        order="nested",
    )


@pytest.mark.benchmark(group="lonlat_to_healpix")
def test_lonlat_to_healpix(benchmark):
    size = 10000
    depth = 12
    lon = Longitude(np.random.rand(size) * 360, u.deg)
    lat = Latitude(np.random.rand(size) * 180 - 90, u.deg)

    benchmark(lonlat_to_healpix, lon=lon, lat=lat, depth=depth, num_threads=1)


@pytest.mark.benchmark(group="lonlat_to_healpix")
def test_lonlat_to_healpix_multithread(benchmark):
    size = 10000
    depth = 12
    lon = Longitude(np.random.rand(size) * 360, u.deg)
    lat = Latitude(np.random.rand(size) * 180 - 90, u.deg)

    benchmark(lonlat_to_healpix, lon=lon, lat=lat, depth=depth)


@pytest.mark.benchmark(group="lonlat_to_healpix")
def test_lonlat_to_healpix_astropy(benchmark):
    size = 10000
    depth = 12
    nside = 1 << depth
    lon = np.random.rand(size) * 360 * u.deg
    lat = (np.random.rand(size) * 180 - 90) * u.deg

    benchmark(
        astropy_healpix.core.lonlat_to_healpix,
        lon=lon,
        lat=lat,
        nside=nside,
        order="nested",
    )


@pytest.mark.benchmark(group="healpix_to_lonlat_broadcast")
def test_healpix_to_lonlat_broadcast(benchmark):
    size = 1000
    depth = np.random.randint(low=0, high=29, size=10)
    ipixels = np.random.randint(12 * 4 ** np.min(depth), size=size)

    benchmark(
        healpix_to_lonlat,
        ipix=ipixels[:, np.newaxis],
        depth=depth[np.newaxis, :],
        num_threads=1,
    )


@pytest.mark.benchmark(group="healpix_to_lonlat_broadcast")
def test_healpix_to_lonlat_broadcast_multithread(benchmark):
    size = 1000
    depth = np.random.randint(low=0, high=29, size=10)
    ipixels = np.random.randint(12 * 4 ** np.min(depth), size=size)

    benchmark(
        healpix_to_lonlat, ipix=ipixels[:, np.newaxis], depth=depth[np.newaxis, :]
    )


@pytest.mark.benchmark(group="healpix_to_lonlat_broadcast")
def test_healpix_to_lonlat_astropy_broadcast(benchmark):
    size = 1000
    depth = np.random.randint(low=0, high=29, size=10)
    nside = 2**depth
    ipixels = np.random.randint(12 * 4 ** np.min(depth), size=size)

    benchmark(
        astropy_healpix.core.healpix_to_lonlat,
        healpix_index=ipixels[:, np.newaxis],
        nside=nside[np.newaxis, :],
        order="nested",
    )


@pytest.mark.benchmark(group="healpix_to_lonlat")
def test_healpix_to_lonlat(benchmark):
    size = 10000
    depth = 12
    ipixels = np.random.randint(12 * 4 ** (depth), size=size)

    benchmark(healpix_to_lonlat, ipix=ipixels, depth=depth, num_threads=1)


@pytest.mark.benchmark(group="healpix_to_lonlat")
def test_healpix_to_lonlat_multithread(benchmark):
    size = 10000
    depth = 12
    ipixels = np.random.randint(12 * 4 ** (depth), size=size)

    benchmark(healpix_to_lonlat, ipix=ipixels, depth=depth)


@pytest.mark.benchmark(group="healpix_to_lonlat")
def test_healpix_to_lonlat_astropy(benchmark):
    size = 10000
    depth = 12
    nside = 1 << depth
    ipixels = np.random.randint(12 * 4 ** (depth), size=size)

    benchmark(
        astropy_healpix.core.healpix_to_lonlat,
        healpix_index=ipixels,
        nside=nside,
        order="nested",
    )


def test_healpix_to_skycoord():
    healpix_to_skycoord(ipix=[0, 2, 4], depth=0)


@pytest.mark.benchmark(group="vertices")
def test_healpix_vertices_lonlat(benchmark):
    depth = 12
    size = 100000
    ipixels = np.random.randint(12 * 4 ** (depth), size=size)

    lon, lat = benchmark(vertices, ipix=ipixels, depth=depth, num_threads=1)


@pytest.mark.benchmark(group="vertices")
def test_healpix_vertices_lonlat_multithread(benchmark):
    depth = 12
    size = 100000
    ipixels = np.random.randint(12 * 4 ** (depth), size=size)

    lon, lat = benchmark(vertices, ipix=ipixels, depth=depth)


@pytest.mark.benchmark(group="vertices")
def test_healpix_vertices_lonlat_astropy(benchmark):
    depth = 12
    size = 100000
    ipixels = np.random.randint(12 * 4 ** (depth), size=size)

    lon2, lat2 = benchmark(
        astropy_healpix.core.boundaries_lonlat,
        healpix_index=ipixels,
        nside=(1 << depth),
        step=1,
        order="nested",
    )


@pytest.mark.benchmark(group="neighbours")
def test_healpix_neighbours(benchmark):
    depth = 12
    size = 100000
    ipixels = np.random.randint(12 * 4 ** (depth), size=size)

    benchmark(neighbours, ipix=ipixels, depth=depth)


lon_c = np.random.rand(1)[0] * 360 * u.deg
lat_c = (np.random.rand(1)[0] * 180 - 90) * u.deg
radius_c = np.random.rand(1)[0] * 45 * u.deg
depth_c = 5


@pytest.mark.benchmark(group="cone_search")
def test_cone_search(benchmark):
    benchmark.pedantic(
        cone_search,
        kwargs={
            "lon": Longitude(lon_c),
            "lat": Latitude(lat_c),
            "radius": radius_c,
            "depth": depth_c,
        },
        iterations=10,
        rounds=100,
    )


@pytest.mark.benchmark(group="cone_search")
def test_cone_search_astropy(benchmark):
    nside = 1 << depth_c
    benchmark.pedantic(
        astropy_healpix.core.healpix_cone_search,
        kwargs={
            "lon": lon_c,
            "lat": lat_c,
            "radius": radius_c,
            "nside": nside,
            "order": "nested",
        },
        iterations=10,
        rounds=100,
    )
